// combined/arranged some fundamental WiFi functions for easy use and for advanced tailor by oneself.
// it is supposed to be as the codebase. Lib form. All materials/techs are from the OpenSource World. arranged by ken woo. v.0.1.
//
// * version 0.1. 20210401. initial version.
//   components: 1) WiFi AP, 2) WiFi STA, 3) http server, 4) http client, 5) mdns with 5.1) addservice and
//   5.2) queryservice and 5.3) auto assigns hostname. 6) system info
//   7) fundamental http urls functions and responses 8) some home-made enhanced wrapper functions.
//
// # dev NOTE: the strange arranged exchanging .cpp and .h contexts could be the only way success for my compiled. no idea.
// # dev NOTE: AP_STA mode for what well procedural controlled is not easy to defined/warpped and omitted.
// # dev NOTE: cMyPing might not ok(the unfriendly native ping functions), better to find a replacement one;
//   it will be blocked at destructor if native ping could not response as expect.
//   current known issue: if more than one instances ping the same ip address at a period, blocked happened.
//
// # NOTE: my_setup_http_server() and my_http_request() have to be impled by user; see the example.
//
// % to do: 1) https. 2) unsure that mdns can announce multiple hostnames for a device. 3) replace ping. 4) A & B http each other problem.


#ifndef _ESP8266_MY_WIFI_UNIFIED_H_
#define _ESP8266_MY_WIFI_UNIFIED_H_


#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <ESP8266HTTPUpdateServer.h>
#include <ESP8266WiFiMulti.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266httpUpdate.h>
#include <LittleFS.h>
#include <ping.h>

#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_

// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
/*// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
*/


// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 0
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
    {"etc", "etc"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "http"
#define LPSERVICE_1 80

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(150) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(50) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 4000 // the timeout during a GET, 4000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 750 // IoT polling at each 750ms
#define TMO_AP_C_CONN_PERIOD 30000 // each 30 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 10000 // each 10 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD 7200000 // each 20 minutes to find internet update
#define TMO_MDNS_Q_PERIOD 30000 // each 30 seconds to query services
#define TMO_PING_PERIOD 5000 // each 5 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 1 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
//#define USE_MDNS_STANDALONE // no auto hostname
#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
//#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // if 1


// ************************** IoT END .h ******************************


#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_


#endif // _ESP8266_MY_WIFI_UNIFIED_H_
