// * version 0.1. 20210401. initial version.
#include"Esp8266MyWiFiUnified.cpp"


// ************************** IoT BEGIN .cpp ******************************
// ========================== global vars ==============================
// -------------------------- general ------------------------------
// for http GET returned
String http_got_str="";

// for connected remote ap ip, in STA mode
char remote_ap_ip[16] = {0};

// for retrieved remote station ip, in AP mode
char remote_sta_ip[16] = {0};

// for remote http server ip
char remote_httph_ip[16] = {0}; // used for remote upd

// remote http server ip, for doing http requests
char http_request_ip[16] = {0};

ESP8266WebServer *server_p=0; // for http server. every device does have.
ESP8266HTTPUpdateServer httpUpdater; // webupdater, will be http://HOSTNAME.local/update. every device does have.

#if defined(USE_STA_WIFI_MULTI)
ESP8266WiFiMulti *WiFiMulti=0;
#endif

#if STA_MODE
bool is_station_mode_established=false;
#endif

#if defined(USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING)
unsigned http_conn_request_state=0; // 0: idle, 1: nak, 2: request ok, 3: receiving data.
int http_conn_request_httpcode=0;
int http_conn_request_data_remaining_sz=0;
WiFiClient *http_conn_WC_client=0;
HTTPClient *http_conn_HC_http=0; // must be declared after WiFiClient for correct destruction order, because used by http.begin(client,...)
#endif

// -------------------------- for ap/sta credential ------------------------------
#if 0 // should be in .cpp file rather than in .h.
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
    {"etc", "etc"},
};
#endif // if 0

// -------------------------- for mdns ------------------------------
#if defined(USE_MDNS_FULL)
// define the services smDnsRecord for this device provides.
// at least the 1st service name=LSERVICE_1/port=LPSERVICE_1 of this device has to be unique, since it is used for part of hostname.
// however if devices have the same hostnames, which will be altered by mdns for uniqueness.
// these service name strings as well as their ports, are defined and used for mdns addServices,
// so as to query all the services within LAN at rumtime. so, services are known by other devices who need it.
// as for the HOSTNAME, will become HOSTNAME_BASE+LSERVICE_1+runtime-altered.

#if 0 // should be in .cpp file rather than in .h.
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // if 0

unsigned mdns_ss_len=sizeof(mdns_ss)/sizeof(smDnsRecord);
#endif // USE_MDNS_FULL

#if 0 // should be in .cpp file rather than in .h.
const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // if 0

unsigned mdns_sskn_len=sizeof(mdns_sskn)/sizeof(smDnsRecord);

String my_hostname=""; // the final local hostname assigned by mDNS.
bool mdns_enabled=false; // set true when connected. could be disabled once aimed and got the ip.
unsigned mDNS_state=0; // 0, disabled(o); 1, active; 2, suspend.


// ========================== IoT basics ==============================
const char * const RST_REASONS[] =
{
    "REASON_DEFAULT_RST",
    "REASON_WDT_RST",
    "REASON_EXCEPTION_RST",
    "REASON_SOFT_WDT_RST",
    "REASON_SOFT_RESTART",
    "REASON_DEEP_SLEEP_AWAKE",
    "REASON_EXT_SYS_RST"
};

const char * const FLASH_SIZE_MAP_NAMES[] =
{
    "FLASH_SIZE_4M_MAP_256_256",
    "FLASH_SIZE_2M",
    "FLASH_SIZE_8M_MAP_512_512",
    "FLASH_SIZE_16M_MAP_512_512",
    "FLASH_SIZE_32M_MAP_512_512",
    "FLASH_SIZE_16M_MAP_1024_1024",
    "FLASH_SIZE_32M_MAP_1024_1024"
};

const char * const OP_MODE_NAMES[] 
{
    "NULL_MODE",
    "STATION_MODE",
    "SOFTAP_MODE",
    "STATIONAP_MODE"
};

const char * const AUTH_MODE_NAMES[] 
{
    "AUTH_OPEN",
    "AUTH_WEP",
    "AUTH_WPA_PSK",
    "AUTH_WPA2_PSK",
    "AUTH_WPA_WPA2_PSK",
    "AUTH_MAX"
};

const char * const PHY_MODE_NAMES[]
{
    "",
    "PHY_MODE_11B",
    "PHY_MODE_11G",
    "PHY_MODE_11N"
};

const char * const EVENT_NAMES[]
{
    "EVENT_STAMODE_CONNECTED",
    "EVENT_STAMODE_DISCONNECTED",
    "EVENT_STAMODE_AUTHMODE_CHANGE",
    "EVENT_STAMODE_GOT_IP",
    "EVENT_SOFTAPMODE_STACONNECTED",
    "EVENT_SOFTAPMODE_STADISCONNECTED",
    "EVENT_MAX"
};

const char * const EVENT_REASONS[]
{
    "",
    "REASON_UNSPECIFIED",
    "REASON_AUTH_EXPIRE",
    "REASON_AUTH_LEAVE",
    "REASON_ASSOC_EXPIRE",
    "REASON_ASSOC_TOOMANY",
    "REASON_NOT_AUTHED",
    "REASON_NOT_ASSOCED",
    "REASON_ASSOC_LEAVE",
    "REASON_ASSOC_NOT_AUTHED",
    "REASON_DISASSOC_PWRCAP_BAD",
    "REASON_DISASSOC_SUPCHAN_BAD",
    "REASON_IE_INVALID",
    "REASON_MIC_FAILURE",
    "REASON_4WAY_HANDSHAKE_TIMEOUT",
    "REASON_GROUP_KEY_UPDATE_TIMEOUT",
    "REASON_IE_IN_4WAY_DIFFERS",
    "REASON_GROUP_CIPHER_INVALID",
    "REASON_PAIRWISE_CIPHER_INVALID",
    "REASON_AKMP_INVALID",
    "REASON_UNSUPP_RSN_IE_VERSION",
    "REASON_INVALID_RSN_IE_CAP",
    "REASON_802_1X_AUTH_FAILED",
    "REASON_CIPHER_SUITE_REJECTED",
};


void GL_print_system_info(Stream & consolePort)
{
    const rst_info * resetInfo = system_get_rst_info();
    consolePort.print(F("system_get_rst_info() reset reason: "));
    consolePort.println(RST_REASONS[resetInfo->reason]);

    consolePort.print(F("system_get_free_heap_size(): "));
    consolePort.println(system_get_free_heap_size());

    consolePort.print(F("system_get_os_print(): "));
    consolePort.println(system_get_os_print());
    system_set_os_print(1);
    consolePort.print(F("system_get_os_print(): "));
    consolePort.println(system_get_os_print());

    consolePort.print(F("system_print_meminfo(): "));
    consolePort.println(system_get_os_print());
    system_print_meminfo();

    consolePort.print(F("system_get_chip_id(): 0x"));
    consolePort.println(system_get_chip_id(), HEX);

    consolePort.print(F("system_get_sdk_version(): "));
    consolePort.println(system_get_sdk_version());

    consolePort.print(F("system_get_boot_version(): "));
    consolePort.println(system_get_boot_version());

    consolePort.print(F("system_get_userbin_addr(): 0x"));
    consolePort.println(system_get_userbin_addr(), HEX);

    consolePort.print(F("system_get_boot_mode(): "));
    consolePort.println(system_get_boot_mode() == 0 ? F("SYS_BOOT_ENHANCE_MODE") : F("SYS_BOOT_NORMAL_MODE"));

    consolePort.print(F("system_get_cpu_freq(): "));
    consolePort.println(system_get_cpu_freq());

    consolePort.print(F("spi_flash_get_id(): 0x"));
    consolePort.println(spi_flash_get_id(), HEX);

    consolePort.print(F("system_get_flash_size_map(): "));
    consolePort.println(FLASH_SIZE_MAP_NAMES[system_get_flash_size_map()]);
}

void GL_pre_setup(){

    GL_print_system_info(Serial);

    // Try pushing frequency to 160MHz.
    system_update_cpu_freq(SYS_CPU_160MHZ);

    // System usually boots up in about 200ms.
    Serial.print(F("system_get_time(): "));
    Serial.println(system_get_time());

    Serial.print(F("system_get_cpu_freq(): "));
    Serial.println(system_get_cpu_freq());

    Serial.print(F("wifi_get_opmode(): "));
    Serial.print(wifi_get_opmode());
    Serial.print(F(" - "));
    Serial.println(OP_MODE_NAMES[wifi_get_opmode()]);

    Serial.print(F("wifi_get_opmode_default(): "));
    Serial.print(wifi_get_opmode_default());
    Serial.print(F(" - "));
    Serial.println(OP_MODE_NAMES[wifi_get_opmode_default()]);

    Serial.print(F("wifi_get_broadcast_if(): "));
    Serial.println(wifi_get_broadcast_if());

    Serial.printf("getResetReason: %s\r\n",         ESP.getResetReason().c_str());
    Serial.printf("getFreeHeap: %d\r\n",            ESP.getFreeHeap());
    Serial.printf("getChipId: %d\r\n",              ESP.getChipId());
    Serial.printf("getCoreVersion: %s\r\n",         ESP.getCoreVersion().c_str());
    Serial.printf("getSdkVersion: %s\r\n",          ESP.getSdkVersion());
    Serial.printf("getCpuFreqMHz: %u\r\n",          ESP.getCpuFreqMHz());
    Serial.printf("getSketchSize: %u\r\n",          ESP.getSketchSize());
    Serial.printf("getFreeSketchSpace: %u\r\n",     ESP.getFreeSketchSpace());
    Serial.printf("getSketchMD5: %s\r\n",           ESP.getSketchMD5().c_str());
    Serial.printf("getFlashChipId: %d\r\n",         ESP.getFlashChipId());
    Serial.printf("getFlashChipSize: %u\r\n",       ESP.getFlashChipSize());
    Serial.printf("getFlashChipRealSize: %u\r\n",   ESP.getFlashChipRealSize());
    Serial.printf("getFlashChipSpeed: %u\r\n",      ESP.getFlashChipSpeed());
    Serial.printf("getCycleCount: %u\r\n",          ESP.getCycleCount());
    ////Serial.printf("getVcc: %d\r\n",                 ESP.getVcc());
}

void GL_print_wifi_general(Stream &consolePort)
{
    consolePort.print(F("wifi_get_channel(): "));
    consolePort.println(wifi_get_channel());
    
    consolePort.print(F("wifi_get_phy_mode(): "));
    consolePort.println(PHY_MODE_NAMES[wifi_get_phy_mode()]);
}


// -------------------------- AP mode ------------------------------
// callback function for wifi auto-messages
void AP_wifi_event_handler_cb(System_Event_t * event)
{
    if (int(event->event)>=7) return;

    if (event->event<sizeof(EVENT_NAMES)/sizeof(EVENT_NAMES[0]))
        Serial.print(EVENT_NAMES[event->event]);
    Serial.print(" [wifi_auto_info] (");

    switch (event->event)
    {
        case EVENT_STAMODE_CONNECTED:
            break;
        case EVENT_STAMODE_DISCONNECTED:
            break;
        case EVENT_STAMODE_AUTHMODE_CHANGE:
            break;
        case EVENT_STAMODE_GOT_IP:
            break;
        case EVENT_SOFTAPMODE_STACONNECTED:
        case EVENT_SOFTAPMODE_STADISCONNECTED:
            {
                char mac[32] = {0};
                snprintf(mac, 32, MACSTR ", aid: %d" , MAC2STR(event->event_info.sta_connected.mac), event->event_info.sta_connected.aid);
                Serial.print(mac);
            }
            break;
        default:
            Serial.print(int(event->event));
    }

    Serial.println(")");
}

void AP_print_softap_config(Stream &consolePort, softap_config const &config)
{
    consolePort.println();
    consolePort.println(F("SoftAP Configuration"));
    consolePort.println(F("--------------------"));

    consolePort.print(F("ssid:            "));
    consolePort.println((char *) config.ssid);

    consolePort.print(F("password:        "));
    consolePort.println((char *) config.password);

    consolePort.print(F("ssid_len:        "));
    consolePort.println(config.ssid_len);

    consolePort.print(F("channel:         "));
    consolePort.println(config.channel);

    consolePort.print(F("authmode:        "));
    consolePort.println(AUTH_MODE_NAMES[config.authmode]);

    consolePort.print(F("ssid_hidden:     "));
    consolePort.println(config.ssid_hidden);

    consolePort.print(F("max_connection:  "));
    consolePort.println(config.max_connection);

    consolePort.print(F("beacon_interval: "));
    consolePort.print(config.beacon_interval);
    consolePort.println("ms");

    consolePort.println(F("--------------------"));
    consolePort.println();
}

void AP_secure_softap_config(softap_config *config, const char *ssid, const char *password)
{ // hidden mode
    // it tests if new data size over the storage and truncates.
    size_t ssidLen     = strlen(ssid)     < sizeof(config->ssid)     ? strlen(ssid)     : sizeof(config->ssid);
    size_t passwordLen = strlen(password) < sizeof(config->password) ? strlen(password) : sizeof(config->password);

    memset(config->ssid, 0, sizeof(config->ssid));
    memcpy(config->ssid, ssid, ssidLen);

    memset(config->password, 0, sizeof(config->password));
    memcpy(config->password, password, passwordLen);

    config->ssid_len = ssidLen;
    config->channel  = 1;
    config->authmode = AUTH_WPA2_PSK;
    config->ssid_hidden = 1;
    config->max_connection = 4;
    config->beacon_interval = 1000;
}

void AP_enter_ap_mode(){
    // correct order
    WiFi.mode(WIFI_AP);
    WiFi.softAP(WIFI_CRED[0][0], WIFI_CRED[0][1]);
    WiFi.begin(); // must, to have WiFi.hostname(name) valid.
    WiFi.hostname(HOSTNAME); // set after WiFi.begin().
    IPAddress myIP = WiFi.softAPIP();

    Serial.println("");
    Serial.println("[-= Enter AP Mode =-]");
    Serial.print("AP IP ADDR: ");
    Serial.println(myIP);

    // get the default config and print it.
    softap_config config;
    wifi_softap_get_config(&config);
    /// if want to be more secure(hidden), modify the configurations.
    /// AP_secure_softap_config(&config, WIFI_CRED[0][0], WIFI_CRED[0][1]);
    /// after config is arranged, set it in.
    /// wifi_softap_set_config(&config);
    AP_print_softap_config(Serial, config);

    GL_print_wifi_general(Serial);

    // start dhcps
    wifi_softap_dhcps_start();
    if (wifi_softap_dhcps_status()==DHCP_STARTED) Serial.println(F("DHCP started"));
    else Serial.println(F("DHCP stopped"));

    // get the AP basic info and print it.
    ip_info info;
    wifi_get_ip_info(SOFTAP_IF, &info);

    /// method 1
    /// Serial.printf("AP IP ADDR: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.ip), ip4_addr2(&info.ip), ip4_addr3(&info.ip), ip4_addr4(&info.ip));
    /// Serial.printf("AP GATEWAY: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.gw), ip4_addr2(&info.gw), ip4_addr3(&info.gw), ip4_addr4(&info.gw));
    /// Serial.printf("AP NETMASK: %3d.%3d.%3d.%3d\r\n", ip4_addr1(&info.netmask), ip4_addr2(&info.netmask), ip4_addr3(&info.netmask), ip4_addr4(&info.netmask));

    // method2
    Serial.printf("AP IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip));
    Serial.printf("AP GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw));
    Serial.printf("AP NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask));

    // This doesn't work on an ESP-01.
    // wifi_set_sleep_type(LIGHT_SLEEP_T);
    // Try this dirty little thing.
    // Doesn't work because ESP-01 module doesn't link XPD_DCDC -> RST.
    // ESP.deepSleep(15000);
}

int AP_connected_station_status(){

    ///// this function for count couldn't sync the number of stat_infos retrieved at realtime.
    unsigned char number_client=wifi_softap_get_station_num(); // Count of stations

    struct station_info *stat_info;
    IPAddress address;
    int i;

    Serial.print("Total connected clients: ");
    Serial.println(number_client);

    for (i=0, stat_info=wifi_softap_get_station_info(); stat_info!=NULL && i<number_client; i++){
        address=(stat_info->ip).addr;

        snprintf(remote_sta_ip, 16, "" IPSTR, IP2STR(&(stat_info->ip)));

        Serial.printf("Station %d. ", i+1);
        Serial.print("IP= ");
        Serial.print(address);
        Serial.print(" with MAC= ");
        Serial.print(stat_info->bssid[0],HEX);
        Serial.print(stat_info->bssid[1],HEX);
        Serial.print(stat_info->bssid[2],HEX);
        Serial.print(stat_info->bssid[3],HEX);
        Serial.print(stat_info->bssid[4],HEX);
        Serial.print(stat_info->bssid[5],HEX);
        Serial.println();

        stat_info = STAILQ_NEXT(stat_info, next);
    }
    wifi_softap_free_station_info();
    delay(100);
    return i; // not exactly realtime the number_client
}


// -------------------------- STATION mode ------------------------------
#if defined(USE_STA_NATIVE) // --------------------------
bool STA_enter_station_mode(){ // 12 seconds waiting
    unsigned char u=0;

    // Connect to WiFi network
    Serial.println("");
    Serial.println("[-= Enter Station Mode =-]");
    Serial.print("Connecting to ");
    Serial.print(WIFI_CRED[0][0]);

    WiFi.hostname(HOSTNAME);
    WiFi.mode(WIFI_STA);
    WiFi.begin(WIFI_CRED[0][0], WIFI_CRED[0][1]);

    while (WiFi.status() != WL_CONNECTED) {
        delay(200);
        Serial.print(".");
        if (u++>60) break; // 12 seconds, might not enough
    }
    Serial.println("");
    if (WiFi.status() != WL_CONNECTED){
        Serial.println(" WiFi connection failed...");
        return false;
    }
    Serial.println("WiFi connected");

    // Print the remote MAC address
    uint8_t macAddr[6];
    WiFi.macAddress(macAddr);
    Serial.printf("Remote MAC adr: %02X:%02X:%02X:%02X:%02X:%02X\r\n", macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]);

    ip_info info;
    wifi_get_ip_info(STATION_IF, &info);
    Serial.printf("IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip));
    Serial.printf("GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw));
    Serial.printf("NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask));

    // resolve the remote ip
    uint32 x=*(uint32*)(&info.ip)&*(uint32*)(&info.netmask);
    ip_addr y=*(ip_addr*)(&x);
    snprintf(remote_ap_ip, 16, "%d.%d.%d.%d",\
        ip4_addr1(&y)? ip4_addr1(&y): 1,\
        ip4_addr2(&y)? ip4_addr2(&y): 1,\
        ip4_addr3(&y)? ip4_addr3(&y): 1,\
        ip4_addr4(&y)? ip4_addr4(&y): 1
        );
    Serial.print(F("Remote AP IP: "));
    Serial.println(remote_ap_ip);

    Serial.printf("Default local hostname: %s\r\n", WiFi.hostname().c_str());
    Serial.print("DNS #1, #2 IP: ");
    WiFi.dnsIP().printTo(Serial);
    Serial.print(", ");
    WiFi.dnsIP(1).printTo(Serial);
    Serial.println();
    Serial.printf("RSSI: %d dBm\r\n", WiFi.RSSI());
    WiFi.printDiag(Serial);

    // Print the local IP address
    Serial.println("");
    Serial.print("IP is assigned: ");
    Serial.println(WiFi.localIP());
    return true;
}
#elif defined(USE_STA_WIFI_MULTI) // --------------------------
bool STA_enter_station_mode(){ // 12 seconds waiting
    unsigned char u=0;

    // Connect to WiFi network
    Serial.println("");
    Serial.println("[-= Enter Station Mode =-]");
    Serial.print("WiFi Connecting...");

    // correct order
    WiFi.mode(WIFI_STA);
    WiFi.begin();
    WiFi.hostname(HOSTNAME); // set after WiFi.begin().

    if (!WiFiMulti){
        WiFiMulti=new ESP8266WiFiMulti;
        if (!WiFiMulti) return false;
        for (int i=0; i<(sizeof(WIFI_CRED)/sizeof(WIFI_CRED[0])); i++)
            WiFiMulti->addAP(WIFI_CRED[i][0], WIFI_CRED[i][1]);
    }

    while (WiFiMulti->run() != WL_CONNECTED) {
        delay(200);
        Serial.print(".");
        if (u++>60){ // 12 seconds, might not enough
            Serial.println(" WiFi connection failed...");
            return false;
        }
    }
    Serial.println("WiFi connected");

    // Print the remote MAC address
    uint8_t macAddr[6];
    WiFi.macAddress(macAddr);
    Serial.printf("Remote MAC adr: %02X:%02X:%02X:%02X:%02X:%02X\r\n", macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5]);

    ip_info info;
    wifi_get_ip_info(STATION_IF, &info);
    Serial.printf("IP ADDR: " IPSTR "\r\n", IP2STR(&info.ip));
    Serial.printf("GATEWAY: " IPSTR "\r\n", IP2STR(&info.gw));
    Serial.printf("NETMASK: " IPSTR "\r\n", IP2STR(&info.netmask));

    // resolve the remote ip
    uint32 x=*(uint32*)(&info.ip)&*(uint32*)(&info.netmask);
    ip_addr y=*(ip_addr*)(&x);
    snprintf(remote_ap_ip, 16, "%d.%d.%d.%d",\
        ip4_addr1(&y)? ip4_addr1(&y): 1,\
        ip4_addr2(&y)? ip4_addr2(&y): 1,\
        ip4_addr3(&y)? ip4_addr3(&y): 1,\
        ip4_addr4(&y)? ip4_addr4(&y): 1
        );
    Serial.print(F("Remote AP IP: "));
    Serial.println(remote_ap_ip);

    Serial.printf("Default local hostname: %s\r\n", WiFi.hostname().c_str());
    Serial.print("DNS #1, #2 IP: ");
    WiFi.dnsIP().printTo(Serial);
    Serial.print(", ");
    WiFi.dnsIP(1).printTo(Serial);
    Serial.println();
    Serial.printf("RSSI: %d dBm\r\n", WiFi.RSSI());
    WiFi.printDiag(Serial);
    
    // Print the local IP address
    Serial.println("");
    Serial.print("IP is assigned: ");
    Serial.println(WiFi.localIP());
    return true;
}
#endif // USE_STA_x // --------------------------


// -------------------------- HTTP server ------------------------------
static const uint8_t x509[] PROGMEM = { // The certificate is stored in PMEM
  0x30, 0x82, 0x01, 0x3d, 0x30, 0x81, 0xe8, 0x02, 0x09, 0x00, 0xfe, 0x56,
  0x46, 0xf2, 0x78, 0xc6, 0x51, 0x17, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86,
  0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b, 0x05, 0x00, 0x30, 0x26, 0x31,
  0x10, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x07, 0x45, 0x53,
  0x50, 0x38, 0x32, 0x36, 0x36, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55,
  0x04, 0x03, 0x0c, 0x09, 0x31, 0x32, 0x37, 0x2e, 0x30, 0x2e, 0x30, 0x2e,
  0x31, 0x30, 0x1e, 0x17, 0x0d, 0x31, 0x37, 0x30, 0x33, 0x31, 0x38, 0x31,
  0x34, 0x34, 0x39, 0x31, 0x38, 0x5a, 0x17, 0x0d, 0x33, 0x30, 0x31, 0x31,
  0x32, 0x35, 0x31, 0x34, 0x34, 0x39, 0x31, 0x38, 0x5a, 0x30, 0x26, 0x31,
  0x10, 0x30, 0x0e, 0x06, 0x03, 0x55, 0x04, 0x0a, 0x0c, 0x07, 0x45, 0x53,
  0x50, 0x38, 0x32, 0x36, 0x36, 0x31, 0x12, 0x30, 0x10, 0x06, 0x03, 0x55,
  0x04, 0x03, 0x0c, 0x09, 0x31, 0x32, 0x37, 0x2e, 0x30, 0x2e, 0x30, 0x2e,
  0x31, 0x30, 0x5c, 0x30, 0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7,
  0x0d, 0x01, 0x01, 0x01, 0x05, 0x00, 0x03, 0x4b, 0x00, 0x30, 0x48, 0x02,
  0x41, 0x00, 0xc6, 0x72, 0x6c, 0x12, 0xe1, 0x20, 0x4d, 0x10, 0x0c, 0xf7,
  0x3a, 0x2a, 0x5a, 0x49, 0xe2, 0x2d, 0xc9, 0x7a, 0x63, 0x1d, 0xef, 0xc6,
  0xbb, 0xa3, 0xd6, 0x6f, 0x59, 0xcb, 0xd5, 0xf6, 0xbe, 0x34, 0x83, 0x33,
  0x50, 0x80, 0xec, 0x49, 0x63, 0xbf, 0xee, 0x59, 0x94, 0x67, 0x8b, 0x8d,
  0x81, 0x85, 0x23, 0x24, 0x06, 0x52, 0x76, 0x55, 0x9d, 0x18, 0x09, 0xb3,
  0x3c, 0x10, 0x40, 0x05, 0x01, 0xf3, 0x02, 0x03, 0x01, 0x00, 0x01, 0x30,
  0x0d, 0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x0b,
  0x05, 0x00, 0x03, 0x41, 0x00, 0x69, 0xdc, 0x6c, 0x9b, 0xa7, 0x62, 0x57,
  0x7e, 0x03, 0x01, 0x45, 0xad, 0x9a, 0x83, 0x90, 0x3a, 0xe7, 0xdf, 0xe8,
  0x8f, 0x46, 0x00, 0xd3, 0x5f, 0x2b, 0x0a, 0xde, 0x92, 0x1b, 0xc5, 0x04,
  0xc5, 0xc0, 0x76, 0xf4, 0xf6, 0x08, 0x36, 0x97, 0x27, 0x82, 0xf1, 0x60,
  0x76, 0xc2, 0xcd, 0x67, 0x6c, 0x4b, 0x6c, 0xca, 0xfd, 0x97, 0xfd, 0x33,
  0x9e, 0x12, 0x67, 0x6b, 0x98, 0x7e, 0xd5, 0x80, 0x8f
};

// And so is the key.  These could also be in DRAM
static const uint8_t rsakey[] PROGMEM = {
  0x30, 0x82, 0x01, 0x3a, 0x02, 0x01, 0x00, 0x02, 0x41, 0x00, 0xc6, 0x72,
  0x6c, 0x12, 0xe1, 0x20, 0x4d, 0x10, 0x0c, 0xf7, 0x3a, 0x2a, 0x5a, 0x49,
  0xe2, 0x2d, 0xc9, 0x7a, 0x63, 0x1d, 0xef, 0xc6, 0xbb, 0xa3, 0xd6, 0x6f,
  0x59, 0xcb, 0xd5, 0xf6, 0xbe, 0x34, 0x83, 0x33, 0x50, 0x80, 0xec, 0x49,
  0x63, 0xbf, 0xee, 0x59, 0x94, 0x67, 0x8b, 0x8d, 0x81, 0x85, 0x23, 0x24,
  0x06, 0x52, 0x76, 0x55, 0x9d, 0x18, 0x09, 0xb3, 0x3c, 0x10, 0x40, 0x05,
  0x01, 0xf3, 0x02, 0x03, 0x01, 0x00, 0x01, 0x02, 0x40, 0x35, 0x0b, 0x74,
  0xd3, 0xff, 0x15, 0x51, 0x44, 0x0f, 0x13, 0x2e, 0x9b, 0x0f, 0x93, 0x5c,
  0x3f, 0xfc, 0xf1, 0x17, 0xf9, 0x72, 0x94, 0x5e, 0xa7, 0xc6, 0xb3, 0xf0,
  0xfe, 0xc9, 0x6c, 0xb1, 0x1e, 0x83, 0xb3, 0xc6, 0x45, 0x3a, 0x25, 0x60,
  0x7c, 0x3d, 0x92, 0x7d, 0x53, 0xec, 0x49, 0x8d, 0xb5, 0x45, 0x10, 0x99,
  0x9b, 0xc6, 0x22, 0x3a, 0x68, 0xc7, 0x13, 0x4e, 0xb6, 0x04, 0x61, 0x21,
  0x01, 0x02, 0x21, 0x00, 0xea, 0x8c, 0x21, 0xd4, 0x7f, 0x3f, 0xb6, 0x91,
  0xfa, 0xf8, 0xb9, 0x2d, 0xcb, 0x36, 0x36, 0x02, 0x5f, 0xf0, 0x0c, 0x6e,
  0x87, 0xaa, 0x5c, 0x14, 0xf6, 0x56, 0x8e, 0x12, 0x92, 0x25, 0xde, 0xb3,
  0x02, 0x21, 0x00, 0xd8, 0x99, 0x01, 0xf1, 0x04, 0x0b, 0x98, 0xa3, 0x71,
  0x56, 0x1d, 0xea, 0x6f, 0x45, 0xd1, 0x36, 0x70, 0x76, 0x8b, 0xab, 0x69,
  0x30, 0x58, 0x9c, 0xe0, 0x45, 0x97, 0xe7, 0xb6, 0xb5, 0xef, 0xc1, 0x02,
  0x21, 0x00, 0xa2, 0x01, 0x06, 0xc0, 0xf2, 0xdf, 0xbc, 0x28, 0x1a, 0xb4,
  0xbf, 0x9b, 0x5c, 0xd8, 0x65, 0xf7, 0xbf, 0xf2, 0x5b, 0x73, 0xe0, 0xeb,
  0x0f, 0xcd, 0x3e, 0xd5, 0x4c, 0x2e, 0x91, 0x99, 0xec, 0xb7, 0x02, 0x20,
  0x4b, 0x9d, 0x46, 0xd7, 0x3c, 0x01, 0x4c, 0x5d, 0x2a, 0xb0, 0xd4, 0xaa,
  0xc6, 0x03, 0xca, 0xa0, 0xc5, 0xac, 0x2c, 0xe0, 0x3f, 0x4d, 0x98, 0x71,
  0xd3, 0xbd, 0x97, 0xe5, 0x55, 0x9c, 0xb8, 0x41, 0x02, 0x20, 0x02, 0x42,
  0x9f, 0xd1, 0x06, 0x35, 0x3b, 0x42, 0xf5, 0x64, 0xaf, 0x6d, 0xbf, 0xcd,
  0x2c, 0x3a, 0xcd, 0x0a, 0x9a, 0x4d, 0x7c, 0xad, 0x29, 0xd6, 0x36, 0x57,
  0xd5, 0xdf, 0x34, 0xeb, 0x26, 0x03
};

#if 0
WiFiServer* HTTPH_create_http_server(){
    // Create an instance of the server
    // specify the 80 port to listen on as an argument
    WiFiServer *a=new WiFiServer(80);

    /// WiFiServerSecure *a=new WiFiServerSecure(443);
    /// a->setServerKeyAndCert_P(rsakey, sizeof(rsakey), x509, sizeof(x509));
    /// a->begin();
    
    // Start the server
    a->begin();
    Serial.println("HTTP Server started");
    return a;
}
#endif

ESP8266WebServer* HTTPH_create_http_server(){ // ESP8266WebServer & MDNS & Updater facilities
    // Create an instance of the server
    // specify the 80 port to listen on as an argument
    ESP8266WebServer *a=new ESP8266WebServer(LPSERVICE_1);

    #if defined(USE_MDNS_STANDALONE)
        if (!MDNS.begin(HOSTNAME)) { // Start the mDNS responder
            Serial.println("Error setting up MDNS responder!");
        }
        Serial.println("mDNS responder started");

        my_hostname=HOSTNAME;
    #endif // if defined(USE_MDNS_STANDALONE)

    httpUpdater.setup(a); // will be http://HOSTNAME.local:LPSERVICE_1/update

    // Start the http server
    a->begin();

    #if defined(USE_MDNS_STANDALONE)
    MDNS.addService(LSERVICE_1, "tcp", LPSERVICE_1);
    #endif // if defined(USE_MDNS_STANDALONE)

    Serial.println(String("UPDATE page is http://")+HOSTNAME+".local:"+LPSERVICE_1+"/update");
    Serial.println("HTTP Server started");
    return a;
}


// -------------------------- HTTP client ------------------------------
#if defined(USE_HTTPC_HTTP_CLIENT) // faster but always got nothing // --------------------------
void HTTPC_http_client_connect(const char *url){
    /// #include <ESP8266HTTPClient.h>
    HTTPClient http;

    Serial.print("\r\n[HTTP begin...]\r\n");

    // configure traged server and url
    /// http.begin("https://192.168.1.12/test.html", "7a 9c f4 db 40 d3 62 5a 6e 21 bc 5c cc 66 c8 3e a1 45 59 38"); // HTTPS
    http.begin(url); // HTTP
    Serial.print("[HTTP GET...]\r\n");
    // start connection and send HTTP header
    int httpCode = http.GET();

    // httpCode will be negative on error
    if (httpCode > 0){
        // HTTP header has been send and Server response header has been handled
        Serial.printf("[HTTP GET... code: %d]\r\n", httpCode);

        // file found at server
        if (httpCode == HTTP_CODE_OK){ // the connection may be broken and stuck here
            Serial.println("-= data begin =-");
            Serial.println(http.getString()); // usually has no data return, so failed.
            Serial.println("-= data end =-");
        }
        else Serial.printf("[HTTP GET... failed, error: %s]\r\n", http.errorToString(httpCode).c_str());
    }
    else Serial.printf("[HTTP Protocol failed, error: %s]\r\n", http.errorToString(httpCode).c_str());

    http.end();
    HTTPC_GET_DELAY;
}
#elif defined(USE_HTTPC_WIFI_CLIENT) // slower and usually failed connection // --------------------------
void HTTPC_http_client_connect(const char *url){
    // Use WiFiClient class to create TCP connections
    WiFiClient client;
    int httpPort = 80;
    if (!client.connect(remote_httph_ip, httpPort)) {
        Serial.println("");
        Serial.println("[HTTP Protocol failed, error: 0]");
        Serial.println("");
        return;
    }

    Serial.print("[Requesting URL: [");
    Serial.print(url);
    Serial.println("]");

    // This will send the request to the server
    client.print(String("GET ") + url + " HTTP/1.1\r\n" +\
        "Host: " + remote_httph_ip + "\r\n" +\
        "Connection: close\r\n");

    unsigned long timeout = millis();
    while (client.available() == 0) {
        if (millis() - timeout > HTTPC_CONN_TIMEOUT) {
            Serial.println("[Client Timeout !]");
            client.stop();
            return;
        }
        HTTPC_GET_PROC_DELAY;
    }

    // Read all the lines of the reply from server and print them to Serial
    Serial.println("-= data begin =-");
    while (client.available()){
        String line = client.readStringUntil('\r');
        Serial.println(line);
    }
    Serial.println("-= data begin =-");
    HTTPC_GET_PROC_DELAY;
}
#elif defined(USE_HTTPC_WIFIHTTP_HYBRID) // --------------------------
void HTTPC_http_client_connect(const char *url){
    WiFiClient client;
    HTTPClient http; //must be declared after WiFiClient for correct destruction order, because used by http.begin(client,...)

    Serial.print("\r\n[HTTP begin...]\r\n");

    // configure server and url
    http.begin(client, url);
    /// http.begin(client, remote_httph_ip, 80, url);

    Serial.print("[HTTP GET...]\r\n");
    // start connection and send HTTP header
    int httpCode = http.GET();
    if (httpCode > 0) {
        // HTTP header has been send and Server response header has been handled
        Serial.printf("[HTTP GET... code: %d]\r\n", httpCode);

        // file found at server
        if (httpCode == HTTP_CODE_OK) {

            // get lenght of document (is -1 when Server sends no Content-Length header)
            int len = http.getSize();
            Serial.println("-= data begin =-");

            #if 0 // with API
            Serial.println(http_got_str=http.getString());
            #else // or "by hand"

            // create buffer for read
            uint8_t buff[128] = { 0 };
            http_got_str="";

            // get tcp stream
            WiFiClient * stream = &client;

            // read all data from server
            while (http.connected() && (len > 0 || len == -1)) {
                // read up to 128 byte
                int c = stream->readBytes(buff, std::min((size_t)len, sizeof(buff)-1));
                Serial.printf("[readBytes: %d]\r\n", c);
                if (!c) {
                    Serial.println("[read timeout]");
                    break; // must have or it will stuck here! and then if stuck at GET(), set timeout
                }

                // write it to Serial
                Serial.write(buff, c);
                buff[c]=0;
                http_got_str+=(char*)buff;

                if (len > 0) len -= c;
            }
            #endif // or "by hand"
            Serial.println("-= data end =-");
        }
        Serial.print("[HTTP connection closed or file end.]\r\n");
    }
    else Serial.printf("[HTTP GET... failed, error: %s]\r\n", http.errorToString(httpCode).c_str());
    http.end();
}
#elif defined(USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING) // --------------------------
bool HTTPC_http_client_connect_GetStart(const char *url){ // return true if request accepted.
    if (http_conn_request_state>=3) return false; // 0: idle, 1: nak, 2: request ok, 3: receiving data.
    http_conn_request_state=0;

    http_conn_WC_client=new WiFiClient;
    http_conn_HC_http=new HTTPClient; // must be declared after WiFiClient for correct destruction order, because used by http.begin(client,...)

    printf("\r\n[HTTP begin...]\r\n");

    // configure server and url
    http_conn_HC_http->begin(*http_conn_WC_client, url);
    /// http_conn_HC_http.begin(http_conn_WC_client, remote_httph_ip, 80, url);

    printf("[HTTP GET...]\r\n");
    // start connection and send HTTP header
    http_conn_request_httpcode = http_conn_HC_http->GET();

    if (http_conn_request_httpcode > 0) { // HTTP header has been send and Server response header has been handled
        printf("[HTTP GET... code: %d]\r\n", http_conn_request_httpcode);

        if (http_conn_request_httpcode == HTTP_CODE_OK) { // file found at server

            // get lenght of document (is -1 when Server sends no Content-Length header)
            http_conn_request_data_remaining_sz = http_conn_HC_http->getSize();
            if (!http_conn_request_data_remaining_sz) http_conn_request_state=2;
            else {
                http_conn_request_state=3;
                printf("-= data begin =-");
            }
        }
        else {
            http_conn_request_state=2;
            printf("[HTTP connection closed or file end.]\r\n");
        }
        return true;
    }
    else {
        http_conn_request_state=1;
        printf("[HTTP GET... failed: %s]\r\n", http_conn_HC_http->errorToString(http_conn_request_httpcode).c_str());
    }
    http_conn_HC_http->end();
    delete http_conn_HC_http; http_conn_HC_http=0;
    delete http_conn_WC_client; http_conn_WC_client=0;
    return false;
}

unsigned HTTPC_http_client_connect_GetCont(unsigned char *buff, unsigned bf_sz){ // use while (http_conn_request_data_remaining_sz).
    // return the read length.

    // read data from server
    if (http_conn_HC_http->connected() && (http_conn_request_data_remaining_sz > 0 || http_conn_request_data_remaining_sz == -1)) {

        // read up to bf_sz byte
        int c = http_conn_WC_client->readBytes(buff, std::min((size_t)http_conn_request_data_remaining_sz, bf_sz));

        printf("[readBytes: %d]\r\n", c);
        if (!c) {
            printf("[read timeout or done read the unknown length data]\r\n");

            // done the read
            http_conn_HC_http->end();
            http_conn_request_data_remaining_sz=0;
            http_conn_request_state=0;
            delete http_conn_HC_http; http_conn_HC_http=0;
            delete http_conn_WC_client; http_conn_WC_client=0;
            Serial.println("-= data end =-");
            return 0;
        }

        // write it to Serial
        Serial.write(buff, c);

        if (((unsigned)http_conn_request_data_remaining_sz) > c){
            if (http_conn_request_data_remaining_sz > c)
                http_conn_request_data_remaining_sz -= c;
            return c;
        }

        // done the read
        http_conn_HC_http->end();
        http_conn_request_data_remaining_sz=0;
        http_conn_request_state=0;
        delete http_conn_HC_http; http_conn_HC_http=0;
        delete http_conn_WC_client; http_conn_WC_client=0;
        Serial.println("-= data end =-");
        return c;
    }
    else {
        http_conn_request_state=0;
        http_conn_request_data_remaining_sz=0;
    }
    return 0;
}
#endif // USE_HTTPC_x // --------------------------


// -------------------------- Remote updater ------------------------------
void gbl_update_started() {
  Serial.println("CALLBACK:  HTTP update process started");
}

void gbl_update_finished() {
  Serial.println("CALLBACK:  HTTP update process finished");
}

void gbl_update_progress(int cur, int total) {
  Serial.printf("CALLBACK:  HTTP update process at %d of %d bytes...\n", cur, total);
}

void gbl_update_error(int err) {
  Serial.printf("CALLBACK:  HTTP update fatal error code %d\n", err);
}

void gbl_updater(const char *url) {
    // wait for WiFi connection

    WiFiClient client;

    // The line below is optional. It can be used to blink the LED on the board during flashing
    // The LED will be on during download of one buffer of data from the network. The LED will
    // be off during writing that buffer to flash
    // On a good connection the LED should flash regularly. On a bad connection the LED will be
    // on much longer than it will be off. Other pins than LED_BUILTIN may be used. The second
    // value is used to put the LED on. If the LED is on with HIGH, that value should be passed
    // ESPhttpUpdate.setLedPin(LED_BUILTIN, LOW);

    // Add optional callback notifiers
    ESPhttpUpdate.onStart(gbl_update_started);
    ESPhttpUpdate.onEnd(gbl_update_finished);
    ESPhttpUpdate.onProgress(gbl_update_progress);
    ESPhttpUpdate.onError(gbl_update_error);

    t_httpUpdate_return ret = ESPhttpUpdate.update(client, url);
    // Or:
    // t_httpUpdate_return ret = ESPhttpUpdate.update(client, "server", 80, "file.bin");

    switch (ret) {
        case HTTP_UPDATE_FAILED:
            Serial.printf("HTTP_UPDATE_FAILD Error (%d): %s\r\n", ESPhttpUpdate.getLastError(), ESPhttpUpdate.getLastErrorString().c_str());
            break;
        case HTTP_UPDATE_NO_UPDATES:
            Serial.println("HTTP_UPDATE_NO_UPDATES");
            break;
        case HTTP_UPDATE_OK:
            Serial.println("HTTP_UPDATE_OK");
            break;
    }
}


// -------------------------- mDNS ------------------------------
#if 0
/*LEAmDNS.h
    (c) 2018, LaborEtArs
    Version 0.9 beta
    Some notes (from LaborEtArs, 2018):
    Essentially, this is an rewrite of the original EPS8266 Multicast DNS code (ESP8266mDNS).
    The target of this rewrite was to keep the existing interface as stable as possible while
    adding and extending the supported set of mDNS features.
    A lot of the additions were basicly taken from Erik Ekman's lwIP mdns app code.
    Supported mDNS features (in some cases somewhat limited):
    - Presenting a DNS-SD service to interested observers, eg. a http server by presenting _http._tcp service
    - Support for multi-level compressed names in input; in output only a very simple one-leven full-name compression is implemented
    - Probing host and service domains for uniqueness in the local network
    - Tiebreaking while probing is supported in a very minimalistic way (the 'higher' IP address wins the tiebreak)
    - Announcing available services after successful probing
    - Using fixed service TXT items or
    - Using dynamic service TXT items for presented services (via callback)
    - Remove services (and un-announcing them to the observers by sending goodbye-messages)
    - Static queries for DNS-SD services (creating a fixed answer set after a certain timeout period)
    - Dynamic queries for DNS-SD services with cached and updated answers and user notifications
    Usage:
    In most cases, this implementation should work as a 'drop-in' replacement for the original
    ESP8266 Multicast DNS code. Adjustments to the existing code would only be needed, if some
    of the new features should be used.
    For presenting services:
    In 'setup()':
      Install a callback for the probing of host (and service) domains via 'MDNS.setProbeResultCallback(probeResultCallback, &userData);'
      Register DNS-SD services with 'MDNSResponder::hMDNSService hService = MDNS.addService("MyESP", "http", "tcp", 5000);'
      (Install additional callbacks for the probing of these service domains via 'MDNS.setServiceProbeResultCallback(hService, probeResultCallback, &userData);')
      Add service TXT items with 'MDNS.addServiceTxt(hService, "c#", "1");' or by installing a service TXT callback
      using 'MDNS.setDynamicServiceTxtCallback(dynamicServiceTxtCallback, &userData);' or service specific
      'MDNS.setDynamicServiceTxtCallback(hService, dynamicServiceTxtCallback, &userData);'
      Call MDNS.begin("MyHostname");
    In 'probeResultCallback(MDNSResponder* p_MDNSResponder, const char* p_pcDomain, MDNSResponder:hMDNSService p_hService, bool p_bProbeResult, void* p_pUserdata)':
      Check the probe result and update the host or service domain name if the probe failed
    In 'dynamicServiceTxtCallback(MDNSResponder* p_MDNSResponder, const hMDNSService p_hService, void* p_pUserdata)':
      Add dynamic TXT items by calling 'MDNS.addDynamicServiceTxt(p_hService, "c#", "1");'
    In loop():
      Call 'MDNS.update();'
    For querying services:
    Static:
      Call 'uint32_t u32AnswerCount = MDNS.queryService("http", "tcp");'
      Iterate answers by: 'for (uint32_t u=0; u<u32AnswerCount; ++u) { const char* pHostname = MDNS.answerHostname(u); }'
      You should call MDNS.removeQuery() sometimes later (when the answers are nott needed anymore)
    Dynamic:
      Install a dynamic query by calling 'DNSResponder::hMDNSServiceQuery hServiceQuery = MDNS.installServiceQuery("http", "tcp", serviceQueryCallback, &userData);'
      The callback 'serviceQueryCallback(MDNSResponder* p_MDNSResponder, const hMDNSServiceQuery p_hServiceQuery, uint32_t p_u32AnswerIndex,
                                         enuServiceQueryAnswerType p_ServiceQueryAnswerType, bool p_bSetContent, void* p_pUserdata)'
      is called for any change in the answer set.
      Call 'MDNS.removeServiceQuery(hServiceQuery);' when the answers are not needed anymore
    Reference:
    Used mDNS messages:
    A (0x01):               eg. esp8266.local A OP TTL 123.456.789.012
    AAAA (0x1C):            eg. esp8266.local AAAA OP TTL 1234:5678::90
    PTR (0x0C, srv name):   eg. _http._tcp.local PTR OP TTL MyESP._http._tcp.local
    PTR (0x0C, srv type):   eg. _services._dns-sd._udp.local PTR OP TTL _http._tcp.local
    PTR (0x0C, IP4):        eg. 012.789.456.123.in-addr.arpa PTR OP TTL esp8266.local
    PTR (0x0C, IP6):        eg. 90.0.0.0.0.0.0.0.0.0.0.0.78.56.34.12.ip6.arpa PTR OP TTL esp8266.local
    SRV (0x21):             eg. MyESP._http._tcp.local SRV OP TTL PRIORITY WEIGHT PORT esp8266.local
    TXT (0x10):             eg. MyESP._http._tcp.local TXT OP TTL c#=1
    Some NOT used message types:
    OPT (0x29):             eDNS
    NSEC (0x2F):            DNSSEC
    License (MIT license)*/
#endif

#if defined(USE_MDNS_FULL)
// only use this facility can query services and alter hostname if violated.
// if rewrite the code to installed more than 1 services by using the following facility, which will become problematic.
// (only the data mdns_ss is empty could success). however, we still addServices manually.
char *pcHostDomain=0; // host domain for negociate. auto assigned.
bool bHostDomainConfirmed=false; // Flag to the confirmation done or not yet of the host domain
MDNSResponder::hMDNSService hMDNSService=0; // The handle of the http service in the MDNS responder
MDNSResponder::hMDNSServiceQuery hMDNSServiceQuery=0; // The handle of the 'http.tcp' service query in the MDNS responder

String mDnsHttpResponse(){ // url/mdns
    printf("[HTTP Response from mDNS]\r\n");

    #if AP_MODE
        IPAddress ip = WiFi.softAPIP();
    #else
        IPAddress ip = WiFi.localIP();
    #endif

    String s="<!DOCTYPE HTML><html><header><h2>Hello from http://";
    s += WiFi.hostname() + ".local/mdns at " + ip.toString() + "</h2></header>";
    s += "<br/><body><h3>HTTP services provided are:</h3><br/>";
    s += "<ol>";
    for (auto info: MDNS.answerInfo(hMDNSServiceQuery)){
        s += "<li>";
        s += info.serviceDomain();
        if (info.hostDomainAvailable()){
            s += "<br/>Hostname: ";
            s += String(info.hostDomain());
            s += (info.hostPortAvailable())? (String(":") + info.hostPort()): "";
        }
        if (info.IP4AddressAvailable()){
            s += "<br/>IPv4:";
            for (auto ip4: info.IP4Adresses()) s += (String(" ") + ip4.toString());
        }
        if (info.txtAvailable()){
            s += "<br/>TXT:<br/>";
            for (auto kv: info.keyValues()) s += "&emsp;" + String(kv.first) + ": " + String(kv.second) + "<br/>";
        }
        s += "</li>";
    }
    s += "</ol><body></html>";
    printf("[%s]\r\n", s.c_str());
    return s;
}

void serviceProbeResult(String service_name, const MDNSResponder::hMDNSService p_hMDNSService, bool the_result){
    // this is the final step for mdns setup done. after it, do the addServices if needed.
    printf("MDNSServiceProbeResultCallback: Service %s probe %s.\r\n", service_name.c_str(), (the_result? "success": "failed"));

    my_hostname=service_name;
    WiFi.hostname(service_name); // change the default into evaluated one.
    printf("\r\ndone the evaluated local website link [http://%s.local/].\r\n", my_hostname.c_str());

    // add more services other than LSERVICE_1.
    for (int i=0; i<mdns_ss_len; i++){
        if (MDNS.addService(mdns_ss[i].service, "tcp", mdns_ss[i].port))
            printf("mDns[%s, tcp, %d] installed.\r\n", mdns_ss[i].service, mdns_ss[i].port);
    }
}

void MDNSServiceQueryCallback(MDNSResponder::MDNSServiceInfo serviceInfo, MDNSResponder::AnswerType answerType, bool p_bSetContent){
    String answerInfo;
    switch (answerType){
        case MDNSResponder::AnswerType::ServiceDomain:
            answerInfo = "ServiceDomain " + String(serviceInfo.serviceDomain());
            break;
        case MDNSResponder::AnswerType::HostDomainAndPort:
            answerInfo = "HostDomainAndPort " + String(serviceInfo.hostDomain()) + ":" + String(serviceInfo.hostPort());
            break;
        case MDNSResponder::AnswerType::IP4Address:
            answerInfo = "IPv4Address ";
            for (IPAddress ip: serviceInfo.IP4Adresses()) answerInfo += (String("- ") + ip.toString());
            break;
        case MDNSResponder::AnswerType::Txt:
            answerInfo = "TXT " + String(serviceInfo.strKeyValue());
            for (auto kv: serviceInfo.keyValues()) answerInfo += "\r\nkv: " + String(kv.first) + ": " + String(kv.second);
            break;
        default:
            answerInfo = "Unknown Answertype";
            break;
    }
    Serial.printf("Answer %s %s.\r\n", answerInfo.c_str(), p_bSetContent? "Modified": "Deleted");
}

void hostProbeResult(String p_pcDomainName, bool p_bProbeResult){
    // MDNSHostProbeResultCallback
    // If the domain is free, the host domain is set and the http service is added.
    // If the domain is already used, a new name is created and the probe is restarted via p_pMDNSResponder->setHostname().

    printf("MDNSHostProbeResultCallback: Host domain '%s.local' is %s.\r\n",\
        p_pcDomainName.c_str(), (p_bProbeResult? "free of use": "already used"));

    if (p_bProbeResult){
        // Set station hostname
        if (pcHostDomain) WiFi.hostname(pcHostDomain);

        if (!bHostDomainConfirmed){
            // Hostname free -> not confirmed yet -> setup service
            bHostDomainConfirmed = true;

            if (!hMDNSService){
                // Add an LSERVICE_1.tcp service to port LPSERVICE_1, using the host domain as instance domain
                hMDNSService = MDNS.addService(0, LSERVICE_1, "tcp", LPSERVICE_1);
                if (hMDNSService){
                    MDNS.setServiceProbeResultCallback(hMDNSService, serviceProbeResult);
    
                    // Add some '_http._tcp' protocol specific MDNS service TXT items
                    // See: http://www.dns-sd.org/txtrecords.html#http
                    MDNS.addServiceTxt(hMDNSService, "user", "");
                    MDNS.addServiceTxt(hMDNSService, "password", "");
                    MDNS.addServiceTxt(hMDNSService, "path", "/");
                }
    
                // Install dynamic LSERVICE_1.tcp service query
                if (!hMDNSServiceQuery){
                    hMDNSServiceQuery = MDNS.installServiceQuery(LSERVICE_1, "tcp", MDNSServiceQueryCallback);
                    if (hMDNSServiceQuery) printf("MDNSProbeResultCallback: Service query item %s.tcp service installed.\r\n", LSERVICE_1);
                    else printf("MDNSProbeResultCallback: Failed to install the service query item %s.tcp service\r\n", LSERVICE_1);
                }
            }
        }
    }
    else {
        // Change hostname, use '-' as divider between base name and index
        if (MDNSResponder::indexDomain(pcHostDomain, "-", 0)) MDNS.setHostname(pcHostDomain);
        else printf("MDNSProbeResultCallback: Failed to update hostname\r\n");
    }
}

bool setup_mDnsResponders(){
    pcHostDomain=0; // can only run once, or will change the result hostname each time. so run if necessary.
    // Setup MDNS responders
    MDNS.setHostProbeResultCallback(hostProbeResult);

    // Init the (currently empty) host domain string with HOSTNAME
    if (!MDNSResponder::indexDomain(pcHostDomain, 0, HOSTNAME) || !MDNS.begin(pcHostDomain)){
        printf(" Error setting up MDNS responder, hostname:%s.\r\n", pcHostDomain);
        return false;
    }
    printf("[%s] MDNS responder started.\r\n", pcHostDomain);
    return true;
}
#endif // if defined(USE_MDNS_FULL)

void mDnsQuery(){ // unable to query the provided itselves
    printf("Sending mDNS query\r\n");
    for (int i=0; i<mdns_sskn_len; i++){
        int n = MDNS.queryService(mdns_sskn[i].service, "tcp"); // Send out query for those tcp services
        if (!n) printf("no service [%s] found.\r\n", mdns_sskn[i].service);
        else {
            printf("service(s) found\r\n");
            for (int j = 0; j < n; j++)
                printf("[%s] %d: %s (%s: %d)\r\n",\
                    mdns_sskn[i].service, j+1, MDNS.hostname(j).c_str(), MDNS.IP(j).toString().c_str(), MDNS.port(j));
        }
    }
    printf("mDNS query done\r\n");
}


// -------------------------- wrapped ping ------------------------------
typedef class cMyPing{
    // the native ping function: the interval which is the waiting time in seconds and then to ping.
    // however for the first time, no waiting and get the result.
    // and the times to ping, if 0, would be 4 times.
    //
    // usage: cMyPing("192.168.1.1").Restart().isAlive(); // for a one-shot. however, must concern about destruct.
    // note: if failed unexpectedly, try to increase the delay in Restart().

    #if 0 // prototype
    typedef void (* ping_recv_function)(void* arg, void *pdata);
    typedef void (* ping_sent_function)(void* arg, void *pdata);

    struct ping_option{
        uint32 count;
        uint32 ip;
        uint32 coarse_time;
        ping_recv_function recv_function;
        ping_sent_function sent_function;
        void* reverse;
    };

    struct ping_resp{
        uint32 total_count;
        uint32 resp_time;
        uint32 seqno;
        uint32 timeout_count;
        uint32 bytes;
        uint32 total_bytes;
        uint32 total_time;
        sint8  ping_err;
    };

    ICACHE_FLASH_ATTR void user_ping_recv_cb(void *arg, void *pdata){
        struct ping_resp *ping_resp = (struct ping_resp*)pdata;
        struct ping_option *ping_opt = (struct ping_option*)arg;
        if (ping_resp->ping_err == -1){
            os_printf("ping failed.\r\n");
        }
        else {
            os_printf("ping recv: bytes = %d, time = %d ms.\r\n", ping_resp->bytes, ping_resp->resp_time);
        }
    };

    ICACHE_FLASH_ATTR void user_ping_sent_cb(void *arg, void *pdata){
        os_printf("user ping finish.\r\n"); // seems not called.
    };

    ICACHE_FLASH_ATTR void ping(const char *ping_ip, unsigned count, unsigned interval_s){
        // the interval_s also means timeout of a trial.

        struct ping_option ping_opt;

        ///// ping_opt = (struct ping_option *)os_zalloc(sizeof(struct ping_option));
        for (int i=0; i<sizeof(ping_opt); i++) ((unsigned char*)(&ping_opt))[i]=0;

        ping_opt.count=count; // try to ping how many times
        ping_opt.coarse_time=interval_s; // ping interval
        ping_opt.ip = ipaddr_addr(ping_ip);

        ping_regist_recv(&ping_opt, user_ping_recv_cb);
        ping_regist_sent(&ping_opt, user_ping_sent_cb); // seems not called.

        ping_start(&ping_opt);
    };
    #endif // if 0; prototype

    typedef struct scMyPingObj{
        ping_option po;
        unsigned success_cnt;
        unsigned failure_cnt;
        unsigned min_resp_ms;
        unsigned max_resp_ms;
        unsigned min_bytes;
        int last_time_failure;
    } scMyPingObj;

    public:

    scMyPingObj obj;

    ICACHE_FLASH_ATTR static void user_ping_recv_cb(void *arg, void *pdata){
        struct ping_resp *ping_resp = (struct ping_resp*)pdata;
        scMyPingObj *ping_opt = (scMyPingObj*)arg;

        #if 1 // for ICACHE_FLASH_ATTR, not for IRAM_ATTR since ping_err is not 4byte.
        if (ping_resp->ping_err == -1){
            ping_opt->failure_cnt++;
            ping_opt->last_time_failure=1;
        }
        #else // for IRAM_ATTR
        if (ping_resp->timeout_count!=ping_opt->failure_cnt){ // however, it not work due to timeout_count wrong.
            ping_opt->failure_cnt=ping_resp->timeout_count;
            ping_opt->last_time_failure=1;
        }
        #endif // if 0

        else {
            if (ping_resp->bytes){
                if (!ping_opt->min_bytes) ping_opt->min_bytes=ping_resp->bytes;
                else if (ping_opt->min_bytes>ping_resp->bytes) ping_opt->min_bytes=ping_resp->bytes;
            }
            if (!ping_opt->min_resp_ms){ping_opt->min_resp_ms=ping_opt->max_resp_ms=ping_resp->resp_time;}
            else if (ping_resp->resp_time<ping_opt->min_resp_ms) ping_opt->min_resp_ms=ping_resp->resp_time;
            else if (ping_resp->resp_time>ping_opt->max_resp_ms) ping_opt->max_resp_ms=ping_resp->resp_time;
            ping_opt->success_cnt++;
            ping_opt->last_time_failure=0;
        }
    };

    void reset_vars(){
        obj.min_bytes=\
        obj.success_cnt=\
        obj.failure_cnt=\
        obj.min_resp_ms=\
        obj.max_resp_ms=0;
        obj.last_time_failure=-1;
    };

    bool Set(const char *ping_ip, unsigned interval_s=1, unsigned count=1){ // count=0 means forever. interval_s=0 means 1.
        if (!isDone()) return false;
        reset_vars();
        if (!count) count=-1;
        if (!interval_s) interval_s=1;

        // prepare the ping structure
        ping_option *base=(ping_option*)&(this->obj);
        for (int i=0; i<sizeof(ping_option)/4; i++) ((int*)base)[i]=0; // because IRAM_ATTR
        // what if not multiple of 4? add tail stub in the object.

        this->obj.po.count=count; // try to ping how many times
        this->obj.po.coarse_time=interval_s; // ping interval
        this->obj.po.ip = ipaddr_addr(ping_ip);
        ping_regist_recv(base, (ping_recv_function)(cMyPing::user_ping_recv_cb));

        return true;
    };

    ~cMyPing(){ // must careful wdt-reset here. maintained by user, instance can not destroy before ping done, or ping mad.
        if (!isDone()) delay(obj.po.coarse_time*(obj.po.count-obj.failure_cnt-obj.success_cnt)*1000);
        while (!isDone()){ // there is a case forever waiting that if the target down and submit ping. hence, the delay used.
            delay(3);
        }
    };

    cMyPing(){reset_vars();};

    cMyPing(const char *ping_ip, unsigned interval_s=1, unsigned count=1){
        reset_vars();
        this->Set(ping_ip, interval_s, count);
    };

    cMyPing& Restart(){ // goes only if a new one or done the previous ping. must check isDone() before able to do Restart().
        if (isDone()){
            reset_vars();
            obj.last_time_failure=1;
            ping_start((ping_option*)&obj);
            delay(50); // use delay for the frist time ping having enough time to response in order to other functions behave correct.
        }
        return *this;
    };

    bool wasResponsed(){return obj.success_cnt+obj.failure_cnt;};

    bool isAlive(){return !obj.last_time_failure;};

    bool isDown(){return ((obj.po.count==obj.failure_cnt) || obj.last_time_failure);};

    bool isDone(){
        unsigned x=obj.success_cnt+obj.failure_cnt;
        return ((obj.last_time_failure==-1) || (x==obj.po.count));
    };

    void PrintPresent(){
        printf("\r\ncMyPing::ping_option[count= %d, ip= %s, tmo= %d.]\r\n",\
            obj.po.count, IPAddress(obj.po.ip).toString().c_str(), obj.po.coarse_time);
        printf("cMyPing[wasResponsed= %d, isAlive= %d, isDown= %d, isDone= %d.]\r\n",\
            wasResponsed(), isAlive(), isDown(), isDone());
        printf("cMyPing[resp time[min: %u, max: %u]; count[all: %u, o: %u, x: %u]; bytes[min: %u].]\r\n\r\n",\
            obj.min_resp_ms, obj.max_resp_ms, obj.po.count, obj.success_cnt, obj.failure_cnt, obj.min_bytes);
    };
} cMyPing;


// -------------------------- use outset ------------------------------
bool Timeout(unsigned ms, unsigned &store_start_time_us, bool one_shot=false, bool one_shot_lasting=false){
    // note: the var store_start_time_us must be exclusively used by this function because it could be modified by this function.
    // !!!!
    // especially note that the micros() will wrap around after 4294967296us, which is around 70 minutes.
    // so, do not use timeout value > 35 minutes; or another abnormal case is past time > 35 minutes before checking.
    // that is, avoid from abnormal of the above two is must by user.

    if (one_shot && !store_start_time_us) return one_shot_lasting;

    ///// if (micros()>=(ms*1000+store_start_time_us)){ ///// this is not ok, will abnormal after power-on 70 minutes.
    if (int(micros())-int(ms*1000+store_start_time_us)>=0){ ///// this is ok

        store_start_time_us=one_shot? 0: micros();
        return true;
    }
    return false;
}

String get_content_type(String filename){
    if (filename.endsWith(".htm")) return "text/html";
    else if (filename.endsWith(".html")) return "text/html";
    else if (filename.endsWith(".css")) return "text/css";
    else if (filename.endsWith(".js")) return "application/javascript";
    else if (filename.endsWith(".png")) return "image/png";
    else if (filename.endsWith(".gif")) return "image/gif";
    else if (filename.endsWith(".jpg")) return "image/jpeg";
    else if (filename.endsWith(".ico")) return "image/x-icon";
    else if (filename.endsWith(".xml")) return "text/xml";
    else if (filename.endsWith(".pdf")) return "application/x-pdf";
    else if (filename.endsWith(".zip")) return "application/x-zip";
    else if (filename.endsWith(".gz")) return "application/x-gzip";
    else if (filename.endsWith(".bin")) return "application/octet-stream";
    return "text/plain";
}

unsigned cstring_to_u32ip(const char *cip){
    unsigned u32ip=0;
    char c[16];
    strcpy(c, cip);
    for (int i=0, j=0, k=0; k<4; k++){
        while (c[i]!='.' && c[i]) i++;
        c[i]=0;
        u32ip+=(String(c+j).toInt()<<(k<<3));
        j=++i;
    }
    return u32ip;
}

void page_gpio_low_cb(){
    //// not correct.
    ////String s = "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n<!DOCTYPE HTML>\r\n<html>\r\nLED is now ";
    ////s += "ON\r\n";
    ////s += "</html>\r\n";
    ////server_p->send(200, "text/plain", s.c_str());
    server_p->send(200, "text/html", "<!DOCTYPE HTML>\r\n<html>\r\nLED is now ON\r\n</html>");
    digitalWrite(2, HIGH);
}

void page_gpio_high_cb(){
    server_p->send(200, "text/html", "<!DOCTYPE HTML>\r\n<html>\r\nLED is now OFF\r\n</html>");
    digitalWrite(2, LOW);
}

void page_led_onoff_inst(){
    /// http://<ip address>/led?state=on will turn the led ON
    /// http://<ip address>/led?state=off will turn the led OFF
    server_p->on("/led", []() {
        String state=server_p->arg("state");
        if (state == "on") digitalWrite(2, LOW);
        else if (state == "off") digitalWrite(2, HIGH);
        server_p->send(200, "text/plain", "LED is now " + state);
    });
}

void page_a_bin_cb(){
    LittleFS.begin();
    File file = LittleFS.open((String("/")+UPDFILE).c_str(), "r");
    if (!file) Serial.println((String("/")+UPDFILE+" opened failed.").c_str());
    else {
        server_p->streamFile(file, "application/octet-stream");
        file.close();
    }
    LittleFS.end();
}

void setup_http_server(){
    server_p=HTTPH_create_http_server();

    server_p->on("/", [](){server_p->send(200, "text/plain", "This is an index page.");});
    server_p->on("/gpio/0", page_gpio_low_cb);
    server_p->on("/gpio/1", page_gpio_high_cb);
    page_led_onoff_inst();
    server_p->on("/upd", page_a_bin_cb);

    #if AP_MODE
        server_p->on("/ip", [](){
            server_p->send(200, "text/plain", WiFi.softAPIP().toString().c_str());
        });
    #else // #if AP_MODE
        server_p->on("/ip", [](){
            server_p->send(200, "text/plain", WiFi.localIP().toString().c_str());
        });
    #endif // #if AP_MODE

    // assign the update website ip to this device
    server_p->on("/webupdinfo", []() { // http://<ip address>/webupdinfo?ip=<remote ip address>.
        String ip=server_p->arg("ip");
        if (ip=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            strcpy(remote_httph_ip, ip.c_str());
            server_p->send(200, "text/plain", (String("1. accepted: ")+remote_httph_ip).c_str());
        }
    });

    #if defined(USE_MDNS_FULL)
        server_p->on("/mdns", []() {server_p->send(200, "text/html", mDnsHttpResponse());});
    #elif defined(USE_MDNS_STANDALONE)
        server_p->on("/mdns", []() {server_p->send(200, "text/plain", (String("http://")+HOSTNAME+".local/").c_str());});
    #endif

    extern void my_setup_http_server();
    my_setup_http_server();
}

void remove_http_server(){
    if (server_p) delete server_p;
    server_p=0;
}

void http_request(){
    extern void my_http_request();
    if (*http_request_ip) my_http_request();
}

void setup_IoT(){
    GL_pre_setup();

#if AP_MODE
    AP_enter_ap_mode();
    wifi_set_event_handler_cb(AP_wifi_event_handler_cb);
    mdns_enabled=true;
    setup_http_server();
#endif
}

void loop_IoT(){
    static unsigned tmo1=micros();
    if (!Timeout(TMO_IOT_LOOP_PERIOD, tmo1)) return; // polling at each 750ms

    #if AP_MODE
        static unsigned tmo2=micros();
        if (Timeout(TMO_AP_C_CONN_PERIOD, tmo2)) // each 30 seconds to show connections.
            Serial.printf("Current STA connections: %d\r\n", AP_connected_station_status());
    #elif STA_MODE
        static unsigned tmo2=micros();
        if (Timeout(TMO_STA_CONN_HEALTH_PERIOD, tmo2, true, true)){
            if (!is_station_mode_established){
                mdns_enabled=is_station_mode_established=STA_enter_station_mode();
                if (is_station_mode_established){ // run at here since mdns needs the ip
                    remove_http_server();
                    setup_http_server();
                }
                delay(1500); // 1.5 seconds to next try.
                return;
            }
    
            #if STA_RECOVER_ALIVE // each 10 seconds(TMO_STA_CONN_HEALTH_PERIOD) to check the connection health.
            else { // if ping no response and debounced, set false then reconnect.
                static unsigned tmo21=micros(), ping_deb=0;
                if (Timeout(TMO_PING_PERIOD, tmo21)){
                    // each 5 seconds(TMO_PING_PERIOD) to ping again when failed.
                    // each 10 seconds(TMO_STA_CONN_HEALTH_PERIOD) to ping again when connected.
                    if (!cMyPing(remote_ap_ip).Restart().isAlive()){
                        printf("[ping x]\r\n");
                        if (++ping_deb>=5){ // 5 times failed then reconnect.
                            ping_deb=0;
                            mDNS_state=mdns_enabled=is_station_mode_established=false;
                            printf("[connection lost, reconnecting...]\r\n");
                        }
                    }
                    else { // yes, really connected and alive.
                        tmo2=micros(); // reset the timeout
                        ping_deb=0;
                        printf("[ping o]\r\n");
                    }
                }
            }
            #endif
    
        }
    #endif

    if (mdns_enabled){
        if (!mDNS_state){
            #if defined(USE_MDNS_FULL)
                if (setup_mDnsResponders() || 1) mDNS_state=1; // since it still get new one if failed.
                else mDNS_state=2;
            #else
                mDNS_state=1;
            #endif
        }

        if (mDNS_state==1){
            MDNS.update();

            static unsigned tmo3=micros();
            if (Timeout(TMO_MDNS_Q_PERIOD, tmo3)) mDnsQuery(); // each 30 seconds to query services.
        }
    }

    if (server_p) server_p->handleClient(); // for local http server

    http_request(); // for local http request to remote

    static unsigned tmo5=micros();
    if (Timeout(TMO_WEB_UPDATE_PERIOD, tmo5)){ // each 20 minutes to find internet update.
        Serial.printf("\r\n-= Updating... =-\r\n");
        if (!*remote_httph_ip) gbl_updater((String("http://")+REMOTE_HOSTNAME+".local/upd").c_str());
        else gbl_updater((String("http://")+remote_httph_ip+"/upd").c_str());
    }
}


// ************************** IoT END .cpp ******************************
