// * version 0.1. 20210401. initial version.
#include <ESP8266WiFi.h>

/* the demo:
 * when this esp8266 device "A" automatically got the hostname, e.g., http://esp8266-iot-http-2.local/, then send the url to "A":
 * http://esp8266-iot-http-2.local/httprequestip?ip=192.168.4.2, which would assign it the "192.168.4.2" as another
 * ESP8266 device "B" which is the http server for "A" to do the http requests, example the "led on/off for a period".
 * this same code without any modification can put to several devices with the same
 * procedure above for them to demand led on/off to one of others.
 * note: use http://esp8266-iot-http-2.local/mdns can probe others info other than itself.
 * use http://esp8266-iot-http-2.local/ip can get itself ip.
*/




// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////
// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 0
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "http"
#define LPSERVICE_1 80

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(75) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(50) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 4000 // the timeout during a GET, 4000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_AP_C_CONN_PERIOD 30000 // each 30 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 20000 // each 20 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD 7200000 // each 20 minutes to find internet update
#define TMO_MDNS_Q_PERIOD 60000 // each 60 seconds to query services
#define TMO_PING_PERIOD 3000 // each 3 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 1 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
//#define USE_MDNS_STANDALONE // no auto hostname
#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
//#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"http", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 8080},
};
#endif // if 1


// ************************** IoT END .h ******************************
////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_




void my_setup_http_server(){

    server_p->on("/httprequestip", []() { // http://<ip address>/httprequestip?ip=<remote ip address>.
        String ip=server_p->arg("ip");
        if (ip=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            strcpy(http_request_ip, ip.c_str());
            server_p->send(200, "text/plain", (String("1. accepted: ")+http_request_ip).c_str());
        }
    });

}

void my_http_request(){
    static unsigned tmo1=micros();
    if (Timeout(60000, tmo1, true, true)){ // after 60 seconds, starting to send http request.
        static unsigned tmo2=micros(), tmo3=tmo2;
        if (!Timeout(1000000, tmo2, true, true)){ // the http requests lasting for 1000 seconds then stop.

            static unsigned tmo4=micros(), tmo5=tmo4;
            if (Timeout(300000, tmo4)){ // after starting requests, then after 300 seconds to disable mdns to see diff.
                mdns_enabled=false;
                printf("\r\nmdns disabled.\r\n");
            }

            if (Timeout(500000, tmo5)){ // then another 200 seconds after, enable it again.
                mdns_enabled=true;
                printf("\r\nmdns enabled.\r\n");
            }
            // for the problem "A" makes "B" blink and "B" makes "A" blink:
            // the diff is no diff. the bottleneck seems exactly while "A" reqs to "B" and while "B" reqs to "A". timeout both.
            // a solution may consider about time sync slots.

            if (Timeout(450, tmo3)){ // each 0.45 seconds to send http request.

                #if defined(USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING)
                if (http_conn_request_data_remaining_sz){
                    unsigned char buff[128];
                    int c=HTTPC_http_client_connect_GetCont(buff, 127);
                    buff[c]=0; http_got_str+=(char*)buff;
                }
                else {
                    Serial.println(http_got_str);
                    http_got_str="";
                    static int x=0;
                    String web_link="http://";
                    web_link+=http_request_ip;
                    web_link+="/gpio/";
                    web_link+=x^=1;
                    Serial.println(web_link);
                    //HTTPC_http_client_connect_GetStart(web_link.c_str());
                    if (!HTTPC_http_client_connect_GetStart(web_link.c_str())){
                        delay((random(3)==1)? random(50)+25: (random(3)==1)? random(25)+12: random(6));
                    }
                }
                #else // if defined(USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING)
                static int x=0;
                String web_link="http://";
                web_link+=http_request_ip;
                web_link+="/gpio/";
                web_link+=x^=1;
                Serial.println(web_link);
                HTTPC_http_client_connect(web_link.c_str());
                Serial.println(http_got_str);
                #endif // if defined(USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING)

            }
        }
    }
}


void setup(){
    Serial.begin(115200);
    delay(5000);

    randomSeed(micros()%ESP.getChipId());

    pinMode(2, OUTPUT);

    setup_IoT();
}


void loop(){
    loop_IoT();
}
